###############################
# Solve Schanckenberg RD system by discretisation of the 
# Laplacian operator (method of lines).
# Plot computation and save all in folders by parameter data, 
# figures and computation data; folders should be 
# previously created.
###############################
#
import os      # For issuing commands to the OS.
import sys     # For determining the Python version.
#
# Libraries needed
from matplotlib.pyplot import *
from matplotlib import rcParams
from pylab import *
from scipy import *
from scipy import linalg, optimize
from scipy import sparse
from scipy.optimize import newton_krylov
from scipy.integrate import ode
from random import *
import numdifftools as nd
import numpy as np
from numpy.linalg import det
#
# Set figure ticks directions
rcParams['xtick.direction']='out'
rcParams['ytick.direction']='out'
#
# Define clear command shell
clear = lambda: os.system('clear')
clear()
#
# Print useful information
print '\n +++++++++++++++++++++++++++++++++++++++++++++'
print '\n Executing on: ', os.uname()
print '\n Python version: ', sys.version
print '\n matplotlib version: ', matplotlib.__version__
print '\n +++++++++++++++++++++++++++++++++++++++++++++ \n'
#
# Set time interval and time step
t0 = 0
tf = 1000
ht = 5e-01
#
# Build spatial discretisation and Laplacian matrix  
L     = 5
M     = int(L*100)
x     = linspace(-L,L,M, endpoint=True)
hx    = x[1]-x[0]
N     = 2*M
diags = array([-2,-1,0,1,2])
data  = zeros((5,N))
#
print 'Numerics parameters:'
print 'Cells =', M
print 'ht    =', ht
print 'hx    =', hx
#
# Three-point standard stencil with Neumann BC
dataN         = data
dataN[1:4][:] = 1.0
dataN[1][M-1] = 0
dataN[0][M]   = -hx/2.0
dataN[1][M]   = 2.0*hx
dataN[3][M]   = 0
dataN[3][M-1] = 2.0*hx
dataN[4][M-1] = -hx/2.0
dataN[2][:]   = -2.0
dataN[2][M-1] = -3.0*hx/2.0
dataN[2][M]   = -3.0*hx/2.0
dataN[0][0]   = -hx/2.0
dataN[-1][-1] = -hx/2.0
dataN[1][0]   = 2.0*hx
dataN[2][0]   = -3.0*hx/2.0
dataN[2][-1]  = -3.0*hx/2.0
dataN[3][-1]  = 2.0*hx
NN            = sparse.spdiags(dataN,diags,N,N)
NN            = NN.transpose()
NN            = NN/hx**2
#
# Three-point standard stencil with periodic BC
dataP         = data
dataP[1][:]   = 1.0
dataP[2][:]   = -2.0
dataP[3][:]   = 1.0
dataP[1][M-1] = 0
dataP[3][M]   = 0
PP            = sparse.spdiags(dataP,diags,N,N)
PP            = sparse.csr_matrix(PP)
PP[0,M-1]     = 1.0
PP[M-1,0]     = 1.0
PP[M,N-1]     = 1.0
PP[N-1,M]     = 1.0
PP            = PP/hx**2
#
# Kinetic parameter values
k1 = 0.05
k2 = 1.0
k3 = 0.3
k4 = 2.0
#
alpha = k3*(k4-k2)/(k4+k2)
gamma = 2*k3*k4/(k4+k2)
beta  = k1*(k4+k2)**2/k3**2
D1    = 10.0**(-3)
D2    = 10.0
#
# All data name
name = 'Schna_RD' + str('alpha%0.3f'%alpha) + str('beta%0.3f'%beta)
#
print '\n'
print 'Parameter set:'
print 'alpha =',alpha
print 'gamma =',gamma
print 'beta  =',beta
print 'D1    =',D1
print 'D2    =',D2
#
# Parameter set file
print '\n Saving parameter set... \t'
parname = 'RDpparvalues/' + name + 'PARValSet.txt'
#
filename = str(parname)
fout     = open(filename,'w')
fout.write('\t Numerics.\n')
fout.write(str(M))
fout.write('\t Cells number \n')
fout.write(str(2*L))
fout.write('\t Domain length \n')
fout.write(str(hx))
fout.write('\t hx \n')
fout.write(str(ht))
fout.write('\t Time step \n')
fout.write(str(t0))
fout.write('\t t0 \n')
fout.write(str(tf))
fout.write('\t tf \n\n')
fout.write('\t Kinetic parameter values.\n')
fout.write(str(alpha))
fout.write('\t alpha \n')
fout.write(str(gamma))
fout.write('\t gamma \n')
fout.write(str(beta))
fout.write('\t beta \n\n')
fout.write('\t Diffusivity matrix parameter values.\n')
fout.write(str(D1))
fout.write('\t D1 \n')
fout.write(str(D2))
fout.write('\t D2 \n')
fout.close()
#
print 'Done. File: ', parname
print '\n'
#
# Diffusion matrix
Ddiags        = array([-2,-1,0,1,2])
Ddata         = zeros((5,N))
Ddata[2][0:M] = D1
Ddata[2][M:N] = D2
DD            = sparse.spdiags(Ddata,diags,N,N)
#
# Set equilibrium point
def Nfield(us,vs):
    #
    ustar = (k2+k4)/k3
    vstar = k4/(k1*ustar**2)
    ff    = k1*us**2*vs + 2*k1*ustar*us*vs + k1*vstar*us**2
    #
    return ff
#
# Set kinetics field
def Schnackenberg(state,t=0):
    # unpack the state vector
    u = state[0]
    v = state[1]
    #
    # compute state derivatives
    ud = alpha*u + beta*v + Nfield(u,v)
    vd = -gamma*u - beta*v - Nfield(u,v)
    #
    Field = array([ud,vd])
    #
    return Field
#
# Set reaction-diffusion field
def RD(t,state):
    # unpack the state vector
    u = state[0:M]
    v = state[M:N]
    #
    # compute state derivatives
    Kinetics      = zeros(N)
    Kinetics[0:M] = alpha*u + beta*v + Nfield(u,v)
    Kinetics[M:N] = -gamma*u - beta*v - Nfield(u,v)
    #
    # compute diffusion terms
    diffusion = DD*PP*state  # Periodic BC
    #diffusion = DD*NN*state  # Neumann BC
    # 
    # return the whole field 
    FU = diffusion + Kinetics
    return FU
#
# Compute the equilibrium point and Jacobian evaluated at it
#
fixedPoint = newton_krylov(Schnackenberg, [0, 0])
#
# Jacobian matrix at the fixed point
Jac = nd.Jacobian(Schnackenberg)(fixedPoint)
C1  = trace(Jac)
C2  = det(Jac)
C3  = D2*Jac[0][0] + D1*Jac[1][1]
C4  = C3**2 - 4.0*D1*D2*C2
#
print 'Equilibrium point:'
print '[u_s,v_s]  =',fixedPoint
print '\n Kinetics Jacobian matrix:'
print 'J[u_s,v_s] =\n', Jac
print '\n Turing conditions:'
print 'TC 1: tra(J) =', C1
print 'TC 2: det(J) =', C2
print 'TC 3: D2 J_11 + D1 J_22 =', C3
print 'TC 4: (D2 J_11 + D1 J_22)^2 - 4 D_1 D_2 det(J) =', C4
#
# Define a fire-starter
def pulse(xx,x0):
    spike = exp(-(xx-x0)**2/2e-02)
    return spike
    #
#
# Upload from a previous computation
#uv0  = np.load('RDddata/Schna_RDalpha0.100beta5.000DataRun.npy')
#u0v0 = uv0[1][:,-1]
#
#
# Set initial condition
eps       = 10*rand(2)
u0v0      = zeros(N)
u0v0[0:M] = fixedPoint[0]*ones(M) + eps[0]*pulse(x,-L/8) + eps[1]*pulse(x,L/8)
u0v0[M:N] = fixedPoint[1]*ones(M) + eps[0]*pulse(x,-L/8) + eps[1]*pulse(x,L/8)
#
state0 = u0v0
#
# Call solver
print '\n Computing... \t'
#
# This is for 'stiff' ODE systems
uv = ode(RD)
uv.set_integrator('lsoda',atol=1e-8, rtol=1e-6)
uv.set_initial_value(state0,t0)
#
# Set solution array
ts = []
ss = []
#
while uv.successful() and (uv.t < tf):
    uv.integrate(uv.t + ht)
    ts.append(uv.t)
    ss.append(uv.y)
#
time    = np.vstack(ts)
UVstate = np.vstack(ss)
#
print 'Done. \n'
#
# Components-wise setting
uu = UVstate[:,0:M]
vv = UVstate[:,M:N]
#
print '\n Plotting... \t'
#
# Plotting
#
# Spatio-temporal surfaces (level curves)
fig0  = figure(0)
ax1   = fig0.add_subplot(211,xticks=[])
imag1 = ax1.imshow(uu.T, interpolation='nearest', alpha=0.7,
                  origin='lower', extent = [0,time[-1],x[0],x[-1]],aspect='auto')
fig0.suptitle(r'$u(t,x),\,v(t,x)$',fontsize=22) 
yl   = ylabel(r'$x$')
yl.set_fontsize(22)
fig0.colorbar(imag1,pad=0.01)
#
ax2   = fig0.add_subplot(212)
imag2 = ax2.imshow(vv.T, interpolation='nearest', alpha=0.7,
                  origin='lower', extent = [0,time[-1],x[0],x[-1]],aspect='auto')
fig0.suptitle(r'$u(t,x),\,v(t,x)$',fontsize=22) 
xl   = xlabel(r'$t$')
yl   = ylabel(r'$x$')
xl.set_fontsize(22)
yl.set_fontsize(22)
fig0.colorbar(imag2,pad=0.01)
subplots_adjust(hspace = 0.07)
#
# Save figure
namefig01 = 'RDffigur/' + name + 'Surffig.png'
savefig(namefig01,dpi=200)
#
print 'Done. File: ', namefig01
print '\n'
#
# Plot initial and final profiles
fig1  = figure(1)
bx1   = fig1.add_subplot(211,xticks=[])
imag1 = bx1.plot(x,uu[0,:],'.-',label=r'$u(0,x)$')
imag1 = bx1.plot(x,uu[-1,:],'.-',label=r'$u(t_\infty,x)$')
yl    = ylabel(r'$u$')
leg1  = bx1.legend(loc='best',fontsize=12,fancybox='True')
yl.set_fontsize(22)
#
bx2   = fig1.add_subplot(212)
imag2 = bx2.plot(x,vv[0,:],'.-',label=r'$v(0,x)$')
imag2 = bx2.plot(x,vv[-1,:],'.-',label=r'$v(t_\infty,x)$')
xl    = xlabel(r'$x$')
yl    = ylabel(r'$v$')
leg2  = bx2.legend(loc='best',fontsize=12,fancybox='True')
xl.set_fontsize(22)
yl.set_fontsize(22)
subplots_adjust(hspace = 0.07)
#
# Save fig
namefig02 = 'RDffigur/' + name + 'Proffig.png'
savefig(namefig02,dpi=200)
#
print 'Done. File: ', namefig02
print '\n'
#
# Save data
print 'Saving data... \t'
#
TUV = []
TUV.append(time)
TUV.append(UVstate.T)
#
dataname = 'RDddata/' + name + 'DataRun'
save(dataname,TUV)
print 'Done. File: ', dataname
print '\n'
#
draw()
#close()
show()
#
